import inspect
import os
from typing import List, Dict

from xumes.core.step import Step

import inspect
import os
import dill


def create_registry_content():
    registry = {}

    def registrar(content: str):
        def decorator(func):
            file_path = inspect.getsourcefile(func)
            file_name = os.path.basename(file_path[:-3])
            if file_name not in registry:
                registry[file_name] = [Step(func, content)]
            else:
                registry[file_name].append(Step(func, content))
            return func

        return decorator

    registrar.all = registry
    return registrar


def exec_registry_function(registry: List[Step], game_context, scenario_name: str):
    """
    Execute all functions from registry
    :param scenario_name:
    :param registry:  of functions
    :param game_context: where the functions will be executed
    """
    return_value = {}
    for r in registry:
        return_value[r] = r.func(game_context, **r.params[scenario_name])
    return return_value


def get_content_from_registry(registry: List[Step]):
    """
    Get all content from registry
    :param registry:  we want to get the content
    """
    return [step.content for step in registry]


def create_registry():
    registry = {}

    def registrar(func):
        file_path = inspect.getsourcefile(func)
        file_name = os.path.basename(file_path[:-3])
        registry[file_name] = func
        return func

    registrar.all = registry
    return registrar



from abc import abstractmethod
from typing import Dict, final

from xumes.entity.i_state_entity import IStateEntity
from xumes.entity.game_element_state import GameElementState
from xumes.entity.i_game_element_state_converter import IGameElementStateConverter


class EntityRegistry(IStateEntity):
    """
       The `EntityRegistry` class is responsible for managing game entities and updating the game state.

       Attributes:
           _game_element_state_converter: An object implementing the `IGameElementStateConverter` interface for converting game elements to states.
           _entities: A dictionary mapping entity names to their corresponding `IStateEntity` instances.
           _game_state: The current game state.

       Methods:
           update(state): Updates the game state when received.
           entity: Property representing the current game state.
           get_all(): Retrieves all game observable entities.
           get(name): Retrieves an entity by its name.
           _update(game_element_state): Updates an existing game entity with the provided `game_element_state`.
           _add(game_element_state): Adds a new game entity with the provided `game_element_state`.
           convert(state_wrapper): Converts a game element, updates or adds it to the state.
           build_entity(game_element_state): Builds a game entity based on the provided `game_element_state`.
    """

    def __init__(self,
                 game_element_state_converter: IGameElementStateConverter
                 ):
        self._game_element_state_converter = game_element_state_converter
        self._entities: Dict[str, IStateEntity] = {
            "test_automation": self
        }
        self._game_state: str = ""

    def update_state(self, state) -> None:
        """
        Update the game state when receive it.
        :param state: game state (ex: "alive", "dead").
        """
        self._game_state = state["state"]

    @property
    @final
    def game_state(self):
        return self._game_state

    @final
    def get_all(self):
        """
        Get every game observables entities.
        :return: every entity
        """
        return self._entities

    @final
    def get(self, name: str):
        return self._entities[name]

    @final
    def _update(self, game_element_state: GameElementState):
        u = self.get(game_element_state.name).update_state(game_element_state.state)
        if u is not None:
            self._entities[game_element_state.name] = u

    @final
    def _add(self, game_element_state: GameElementState):
        self._entities[game_element_state.name] = self.build_entity(game_element_state)

    @final
    def convert(self, state_wrapper):
        """
        Convert the game element, and update or add it in the state.
        :param state_wrapper: game element from the list receive.
        """
        ges = self._game_element_state_converter.convert(state_wrapper)
        if ges.name in self._entities:
            self._update(ges)
        else:
            self._add(ges)

    @abstractmethod
    def build_entity(self, game_element_state: GameElementState) -> IStateEntity:
        """
        Implements a way to redirect the building of game entities.
        Need to call build method from the implemented game entities.
        :param game_element_state:
        """
        raise NotImplementedError


def choose_delegate(value):
    if isinstance(value, bool):
        return EntityBoolAdapter(value)
    elif isinstance(value, int):
        return EntityIntAdapter(value)
    elif isinstance(value, float):
        return EntityFloatAdapter(value)
    elif isinstance(value, complex):
        return EntityComplexAdapter(value)
    elif isinstance(value, str):
        return EntityStrAdapter(value)
    elif isinstance(value, list):
        return EntityListAdapter(value)
    elif isinstance(value, tuple):
        return EntityTupleAdapter(value)
    elif isinstance(value, set):
        return EntitySetAdapter(value)
    elif isinstance(value, dict):
        if "__type__" in value:
            return EntityObject(value)
        return EntityDictAdapter(value)


class Entity(IStateEntity):

    def __hash__(self):
        raise NotImplementedError


class EntityDictAdapter(dict, Entity):

    def __init__(self, value):
        dict.__init__({})
        for k in value:
            if k != "__type__":
                if isinstance(value[k], Entity):
                    self[k] = value[k]
                else:
                    self[k] = choose_delegate(value[k])

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityDictAdapter(state)

    def update_state(self, state):
        for k in state:
            if k in self:
                self[k] = self[k].update_state(state[k])
            else:
                self[k] = choose_delegate(state[k])
        return self

    def __hash__(self):
        return hash(tuple(self.keys()))


class EntityObject(EntityDictAdapter):

    def __init__(self, value):
        EntityDictAdapter.__init__(self, value)
        if "__type__" in value:
            self.__type__ = value["__type__"]

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityObject(state)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError(f"'{self.__type__}' object has no attribute '{name}'")

    def update_state(self, state):
        for k in state:
            if k != "__type__":
                if k in self and isinstance(self[k], Entity):
                    self[k] = self[k].update_state(state[k])
                else:
                    self[k] = choose_delegate(state[k])
        return self


class EntityBoolAdapter(int, Entity):

    def __init__(self, value):
        bool.__init__(value)

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityBoolAdapter(state)

    def update_state(self, state):
        return EntityBoolAdapter(state)

    def __hash__(self):
        return hash(bool(self))


class EntityIntAdapter(int, Entity):

    def __init__(self, value):
        int.__init__(value)

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityIntAdapter(state)

    def update_state(self, state):
        return EntityIntAdapter(state)

    def __hash__(self):
        return hash(int(self))


class EntityFloatAdapter(float, Entity):

    def __init__(self, value):
        float.__init__(value)

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityFloatAdapter(state)

    def update_state(self, state):
        return EntityFloatAdapter(state)

    def __hash__(self):
        return hash(float(self))


class EntityComplexAdapter(complex, Entity):

    def __init__(self, value):
        complex.__init__(value)

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityComplexAdapter(state)

    def update_state(self, state):
        return EntityComplexAdapter(state)

    def __hash__(self):
        return hash(complex(self))


class EntityStrAdapter(str, Entity):

    def __init__(self, value):
        str.__init__(value)

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityStrAdapter(state)

    def update_state(self, state):
        return EntityStrAdapter(state)

    def __hash__(self):
        return hash(str(self))


class EntityListAdapter(list, Entity):

    def __init__(self, value):
        list.__init__([])
        for i in value:
            if isinstance(i, Entity):
                self.append(i)
            else:
                self.append(choose_delegate(i))

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityListAdapter(state)

    def update_state(self, state):
        self.clear()
        for i in state:
            if isinstance(i, Entity):
                self.append(i)
            else:
                self.append(choose_delegate(i))
        return self

    def __hash__(self):
        return hash(tuple(self))


class EntityTupleAdapter(tuple, Entity):

    def __init__(self, value):
        tuple.__init__((i if isinstance(i, Entity) else choose_delegate(i) for i in value))

    @staticmethod
    def build(state) -> IStateEntity:
        return EntityTupleAdapter(state)

    def update_state(self, state):
        return EntityTupleAdapter(state)

    def __hash__(self):
        return hash(tuple(self))


class EntitySetAdapter(set, Entity):

    def __init__(self, value):
        super().__init__()
        for i in value:
            self.add(i if isinstance(i, Entity) else choose_delegate(i))

    @staticmethod
    def build(state) -> IStateEntity:
        return EntitySetAdapter(state)

    def update_state(self, state):
        self.clear()
        for i in state:
            self.add(i if isinstance(i, Entity) else choose_delegate(i))
        return self

    def __hash__(self):
        return hash(tuple(self))


class AutoEntityManager(EntityRegistry):
    def build_entity(self, game_element_state: GameElementState) -> IStateEntity:
        if game_element_state.type == "unknown":
            if isinstance(game_element_state.state, int):
                return EntityIntAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, bool):
                return EntityBoolAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, dict):
                return EntityDictAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, str):
                return EntityStrAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, float):
                return EntityFloatAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, list):
                return EntityListAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, tuple):
                return EntityTupleAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, set):
                return EntitySetAdapter.build(game_element_state.state)
            elif isinstance(game_element_state.state, complex):
                return EntityComplexAdapter.build(game_element_state.state)
            else:
                raise ValueError(f"Unknown type {type(game_element_state.state)}")
        return EntityObject.build(game_element_state.state)

    def __getattr__(self, item):
        if item in self._entities:
            return self._entities[item]
        raise AttributeError



from typing import Any, Dict


class IComGameInstance:
    """
      Communication between the test_automation and the game.
      Methods:
          observe: Send the game state to the training server.
          action: Wait for the training server to send an action.
          run: Start the communication service (e.g., start the app of a REST API).
    """

    def push_dict(self, dictionary) -> None:
        """
        Send a dictionary to the game.
        """
        raise NotImplementedError

    def get_dict(self) -> Dict[str, Any]:
        """
        Receive a dictionary from the game.
        """
        raise NotImplementedError

    def get_int(self) -> int:
        """
        Receive an integer from the game.
        """
        raise NotImplementedError

    def init_socket(self, port) -> None:
        """
        Used to start the communication service (using threads).
        For example: start the app of a REST API.
        """
        raise NotImplementedError

    def stop_socket(self) -> None:
        """
        Used to stop the communication service.
        """
        raise NotImplementedError





def parse_json_with_eval(json_obj):
    if isinstance(json_obj, dict):
        for key, value in json_obj.items():
            json_obj[key] = parse_json_with_eval(value)
        return json_obj
    elif isinstance(json_obj, list):
        return [parse_json_with_eval(item) for item in json_obj]
    elif isinstance(json_obj, str):
        try:
            return eval(json_obj)
        except (SyntaxError, NameError, TypeError):
            # Si l'évaluation échoue, on laisse la valeur inchangée
            return json_obj
    else:
        return json_obj



from abc import abstractmethod

from xumes.test_automation.test_runner import TestRunner


class Behavior:

    def __init__(self):
        self._test_runner: TestRunner = None
        self._mode = None

    def set_mode(self, mode: str):
        self._mode = mode

    def set_test_runner(self, test_runner):
        self._test_runner = test_runner

    def test_runner(self):
        return self._test_runner

    @abstractmethod
    def execute(self, feature, scenario):
        """
        Execute the behavior algorithm.
        """
        raise NotImplementedError

    def __getattr__(self, item):
        """
        Retrieves an entity from the game service.
        """
        return self._test_runner.get_entity(item)

    @abstractmethod
    def terminated(self) -> bool:
        """
        Check if the game has terminated.
        """
        raise NotImplementedError



import numpy as np
from scipy.stats import ttest_1samp
from statsmodels.stats.proportion import proportions_ztest


def check_equal_values(sample, pop_mean):
    if all(value == pop_mean for value in sample):
        return True
    else:
        return False


def check_all_equal_values(sample):
    if len(set(sample)) == 1:
        return True
    else:
        return False


def hypothesis_test(sample, pop_mean, alpha):
    # Case if all values are equals
    if check_equal_values(sample, pop_mean):  # if all values are equals to pop_mean
        return None, None, None, None, False
    elif check_all_equal_values(sample):  # if all values are equals
        return None, None, None, None, True

    # Case if all values are not equals

    # Compute Student's t-test
    t_statistic, p_value_t = ttest_1samp(sample, pop_mean)

    # Compute Proportions z-test we need to add 0.00000000001 to avoid division by zero
    count = sum([1 for value in sample if value == pop_mean]) + 0.00000000001
    nobs = len(sample) + 0.00000000001
    _, p_value_prop = proportions_ztest(count, nobs, pop_mean)

    # Compute combined p-value
    p_value_combined = min(p_value_t, p_value_prop) * 2

    # Check if the difference is significant
    if p_value_combined < alpha:
        significant_difference = True
    else:
        significant_difference = False

    return t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference


class IAssertionStrategy:
    """
    Implements the Strategy design pattern to define the interface for assertions
    """

    def test(self, data) -> bool:
        raise NotImplementedError


class Assertion(IAssertionStrategy):
    """
    Abstract class for assertions
    Implements the test method of the IAssertionStrategy interface
    and compute the hypothesis test for the given data
    """

    def __init__(self, value=None):
        self._value = value
        if value is not None:
            self._type = type(value)
        else:
            self._type = None

    def test(self, data) -> bool:
        if self._type is None:
            raise NotImplementedError("Assertion type is not defined")
        if len(data) == 0:
            raise ValueError("Data is empty")
        if not isinstance(data[0], self._type):
            raise TypeError("Data type is not equal to assertion type")

        return False


class AssertionStatistical(Assertion):

    def __init__(self, value=None, alpha=0.001):
        super().__init__(value)
        self._alpha = alpha


class AssertionEqualStatistical(AssertionStatistical):
    """
    Overloads the test method of the Assertion class to test if the mean of the data is equal to the given value
    """

    def test(self, data) -> bool:
        super().test(data)

        value = self._value
        if self._type != float and self._type != int:
            data = np.array([1 if x == self._value else -1 for x in data])
            value = 1

        t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data, value,
                                                                                                         self._alpha)

        return not significant_difference


class AssertionBetweenStatistical(AssertionStatistical):
    """
    Overloads the test method of the Assertion class to test if the mean of the data is between the given values
    """

    def __init__(self, min_value, max_value, alpha=0.001):
        super().__init__(alpha=alpha)

        if min_value > max_value:
            min_value, max_value = max_value, min_value
        elif min_value == max_value:
            raise ValueError("min_value and max_value cannot be equal")

        if type(min_value) != type(max_value):
            raise TypeError("min_value and max_value must have the same type")

        self._min_value = min_value
        self._max_value = max_value
        self._type = type(min_value)

    def test(self, data) -> bool:
        super().test(data)

        data = np.array([1 if self._min_value <= x <= self._max_value else -1 for x in data])
        value = 1

        t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                         value,
                                                                                                         self._alpha)

        if p_value_combined:
            return p_value_combined > self._alpha

        return not significant_difference


class AssertionLessThanStatistical(AssertionStatistical):
    """
    Overloads the test method of the Assertion class to test if the mean of the data is less than the given value
    """

    def test(self, data) -> bool:
        super().test(data)

        if self._type != float and self._type != int:
            data = np.array([1 if x < self._value else -1 for x in data])
            value = 1

            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             value,
                                                                                                             self._alpha)

            return not significant_difference

        else:
            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             self._value,
                                                                                                             self._alpha)

            if p_value_combined and t_statistic:
                return p_value_combined / 2 < self._alpha and t_statistic < 0
            else:
                return np.mean(data) < self._value


class AssertionLessThanOrEqualStatistical(AssertionStatistical):
    """
    Overloads the test method of the Assertion class to test if the mean of the data is less than or equal to the given value
    """

    def test(self, data) -> bool:
        super().test(data)

        if self._type != float and self._type != int:
            data = np.array([1 if x <= self._value else -1 for x in data])
            value = 1

            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             value,
                                                                                                             self._alpha)

            return not significant_difference

        else:
            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             self._value,
                                                                                                             self._alpha)
            if t_statistic:
                return t_statistic < 0
            else:
                return np.mean(data) <= self._value


class AssertionGreaterThanStatistical(AssertionStatistical):
    """
    Overloads the test method of the Assertion class to test if the mean of the data is greater than the given value
    """

    def test(self, data) -> bool:
        super().test(data)

        if self._type != float and self._type != int:
            data = np.array([1 if x > self._value else 0 for x in data])
            value = 1

            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             value,
                                                                                                             self._alpha)

            return not significant_difference

        else:
            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             self._value,
                                                                                                             self._alpha)

            if p_value_combined and t_statistic:
                return p_value_combined / 2 < self._alpha and t_statistic > 0
            else:
                return np.mean(data) > self._value


class AssertionGreaterThanOrEqualStatistical(AssertionStatistical):
    """
    Overloads the test method of the Assertion class to test if the mean of the data is greater than or equal to the given value
    """

    def test(self, data) -> bool:
        super().test(data)

        if self._type != float and self._type != int:
            data = np.array([1 if x >= self._value else -1 for x in data])
            value = 1

            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             value,
                                                                                                             self._alpha)
            return not significant_difference

        else:
            t_statistic, p_value_t, p_value_prop, p_value_combined, significant_difference = hypothesis_test(data,
                                                                                                             self._value,
                                                                                                             self._alpha)
            if t_statistic:
                return t_statistic > 0
            else:
                return np.mean(data) >= self._value


class AssertionEqual(Assertion):

    def test(self, data) -> bool:
        super().test(data)
        return any([value == self._value for value in data])


class AssertionBetween(Assertion):

    def __int__(self, min_value, max_value):
        super().__init__()
        if min_value > max_value:
            min_value, max_value = max_value, min_value
        elif min_value == max_value:
            raise ValueError("min_value and max_value cannot be equal")

        if type(min_value) != type(max_value):
            raise TypeError("min_value and max_value must have the same type")

        self._min_value = min_value
        self._max_value = max_value
        self._type = type(min_value)

    def test(self, data) -> bool:
        super().test(data)
        return any([self._min_value <= value <= self._max_value for value in data])


class AssertionLessThan(Assertion):

    def test(self, data) -> bool:
        super().test(data)
        return any([value < self._value for value in data])


class AssertionLessThanOrEqual(Assertion):

    def test(self, data) -> bool:
        super().test(data)
        return any([value <= self._value for value in data])


class AssertionGreaterThan(Assertion):

    def test(self, data) -> bool:
        super().test(data)
        return any([value > self._value for value in data])


class AssertionGreaterThanOrEqual(Assertion):

    def test(self, data) -> bool:
        super().test(data)
        return any([value >= self._value for value in data])



from xumes.core.registry import create_registry, create_registry_content


config = create_registry()
given = create_registry_content()
when = create_registry_content()
then = create_registry_content()

config_registry = config.all
given_registry = given.all
when_registry = when.all
then_registry = then.all



import logging
from typing import TypeVar, final

from xumes.assertions.assertion_bucket import AssertionBucket
from xumes.core.errors.running_ends_error import RunningEndsError
from xumes.core.modes import TEST_MODE
from xumes.core.registry import exec_registry_function
from xumes.entity.entity_manager import AutoEntityManager
from xumes.entity.implementations.json_impl.json_game_element_state_converter import \
    JsonGameElementStateConverter
from xumes.test_automation.driver import Driver
from xumes.test_automation.game_instance_service import GameInstanceService
from xumes.test_automation.test_context import TestContext

OBST = TypeVar("OBST")


class TestRunner:
    """
    The `TestRunner` class is a central component of Xumes. It manages communication between communication service,
    the execution of the game itself, and external events that can modify the game state.
    """

    def __init__(self, game_instance_service: GameInstanceService, number_max_of_steps: int = None,
                 number_max_of_tests: int = None,
                 mode: str = TEST_MODE, feature_name: str = None,
                 scenario_name: str = None, test_queue=None, steps=None, registry_queue=None):
        """
        The constructor for the GameService class.
        """
        self._game_instance_service: GameInstanceService = game_instance_service
        self._feature = feature_name
        self._scenario = scenario_name
        self._mode = mode
        self._number_of_steps = 0
        self._number_max_of_steps = number_max_of_steps
        self._number_of_tests = 0
        self._number_max_of_tests = number_max_of_tests
        self._steps = steps
        self._entity_manager = AutoEntityManager(JsonGameElementStateConverter())

        self._assertion_bucket = AssertionBucket(test_name=f"{self._feature}/{self._scenario}",
                                                 queue=test_queue)
        self.driver = Driver()

        self._config_registry, self._given_registry, self._when_registry, self._then_registry = registry_queue.get()

        self._context = TestContext(self._entity_manager, self.driver, self._assertion_bucket)

    def run(self, port):
        self._game_instance_service.run(port)

    def config(self):
        self._config_registry[self._steps](self)
        return self.driver()

    def given(self):
        exec_registry_function(registry=self._given_registry[self._steps], game_context=self._context,
                               scenario_name=self._scenario)

    def when(self):
        return exec_registry_function(registry=self._when_registry[self._steps], game_context=self._context,
                                      scenario_name=self._scenario)

    def then(self):
        return exec_registry_function(registry=self._then_registry[self._steps], game_context=self._context,
                                      scenario_name=self._scenario)

    def episode_finished(self) -> bool:
        # when an episode is finished, we collect the assertions
        if self._mode == TEST_MODE:
            if self._number_of_tests >= self._number_max_of_tests:
                self._do_assert()
                raise RunningEndsError
            else:
                try:
                    self.then()
                    self._assertion_bucket.reset_iterator()
                except KeyError:
                    pass
            self._number_of_tests += 1

        return True

    def _do_assert(self) -> None:
        self._assertion_bucket.assertion_mode()
        self.then()
        self._assertion_bucket.send_results()
        self._assertion_bucket.clear()
        self._assertion_bucket.collect_mode()

    @final
    def reset(self):
        """
        Resets the game state by calling the given and reset methods of the TestRunner instance.
        """
        self.given()
        self._game_instance_service.reset()

    @final
    def push_action_and_get_state(self, actions):
        """
        Pushes actions to the TestRunner and retrieves the game state.

        Parameters:
            actions: The actions to be pushed to the TestRunner.
        """
        logging.debug(f"Pushing actions: {actions}")
        methods = self.driver()
        states = self._game_instance_service.push_actions_and_get_state(actions, methods)
        logging.debug(f"Received states: {states}")
        for state in states.items():
            self._entity_manager.convert(state)

    @final
    def finished(self):
        """
        Checks if the game has finished by calling the finish method of the TestRunner instance.

        Returns:
            bool: True if the game has finished, False otherwise.
        """
        return self._game_instance_service.finish()

    @final
    def retrieve_state(self) -> None:
        """
        Retrieves the game state by calling the get_state method of the TestRunner instance.
        """
        # self._test_runner.run_loop()
        states = self._game_instance_service.get_state()
        logging.debug(f"Received states: {states}")
        for state in states.items():
            self._entity_manager.convert(state)

    @final
    @property
    def game_state(self):
        """
        Retrieves the game state by calling the get_entity_manager method of the TestRunner instance.

        Returns:
            The current game state.
        """
        return self._entity_manager

    @final
    def get_entity(self, name):
        return self._entity_manager.get(name)



from typing import List, Dict


class IComGameEngine:

    def start_scenario(self, name: str, methods: List[Dict], fps_limit: int, render: bool):
        raise NotImplementedError

    def stop_scenario(self, port: int):
        raise NotImplementedError

    def start_scenarios(self, name: str, methods: List[Dict], fps_limit: int, render: bool):
        raise NotImplementedError

    def stop_scenarios(self, pids: List[int]):
        raise NotImplementedError



import importlib.util
import sys

import os
from abc import ABC, abstractmethod
from typing import List

from xumes.test_automation.registries import given, when, then


class Scenario:

    def __init__(self, name: str = None, steps: str = None, feature=None, tags=None):
        self.name = name
        self.steps: str = steps
        self.feature: Feature = feature

        if tags is None:
            tags = []
        self.tags = tags


class Feature:

    def __init__(self, scenarios=None, name: str = None):
        if scenarios is None:
            scenarios = []
        self.scenarios: List[Scenario] = scenarios
        self.name = name
        self.processes = []


class FeatureStrategy(ABC):
    """
    FeatureStrategy is a class that implements the strategy pattern to define a way to get
    all features.
    """

    def __init__(self, alpha: float = 0.001, steps_path: str = "./"):
        self.features: List[Feature] = []
        self._steps_files: List[str] = []

        self._alpha = alpha
        self._load_tests(steps_path)

        self.given = given
        self.when = when
        self.then = then

    def _load_tests(self, path: str = "./"):
        current_dir = os.path.dirname(os.path.abspath(__file__))

        parent_dir = os.path.abspath(os.path.join(current_dir, ".."))
        sys.path.insert(0, parent_dir)

        for file in os.listdir(path):
            if file.endswith(".py"):
                module_path = os.path.join(path, file)
                module_path = os.path.abspath(module_path)
                module_name = os.path.basename(module_path)[:-3]

                spec = importlib.util.spec_from_file_location(module_name, module_path)
                module_dep = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module_dep)

                self._steps_files.append(file[:-3])

    @abstractmethod
    def retrieve_feature(self, path: str):
        """
        Get all features.
        """
        raise NotImplementedError


class DummyFeatureStrategy(FeatureStrategy):
    def retrieve_feature(self, path: str):
        pass



import dill


class Step:

    def __init__(self, func, content, params=None):
        self.func = func
        self.content = content
        if params is None:
            params = {}
        self.params = params

    def add_params(self, scenario_name, params):
        self.params[scenario_name] = params

    def __reduce__(self):
        return self.__class__, (self.func, self.content, self.params)



import logging
from abc import ABC, abstractmethod
from typing import TypeVar, List, Optional

from xumes.core.model_helper import trainer_name, model_path
from xumes.core.modes import TEST_MODE, TRAIN_MODE
from xumes.test_automation.behavior import Behavior
from xumes.modules.reinforcement_learning.i_trainer import ITrainer
from xumes.modules.reinforcement_learning.stable_baselines_trainer import StableBaselinesTrainer

OBST = TypeVar("OBST")


class Agent(Behavior, ABC):
    """
    The `Agent` class represents a machine learning agent that interacts with the game.

    Attributes:
        - `_save_path` (str): The path to save the trained model.
        - `_eval_freq` (int): The evaluation frequency during training.
        - `_logs_path` (Optional[str]): The path to save training logs.
        - `_logs_name` (Optional[str]): The name of the log files.
        - `_previous_model_path` (Optional[str]): The path to a pre-trained model, if available.
    """

    def __init__(self, save_path: str = None, eval_freq: int = 1000, logs_path: Optional[str] = None,
                 logs_name: Optional[str] = None, previous_model_path: Optional[str] = None, **kwargs):
        super().__init__()
        self._trainer: ITrainer = StableBaselinesTrainer(self, **kwargs)

        self._save_path = save_path
        self._eval_freq = eval_freq
        self._logs_path = logs_path
        self._logging = False
        self._previous_model_path = previous_model_path

    def get_trainer_info(self):
        return {
            "eval_freq": self._eval_freq,
            "previous_model_path": self._previous_model_path
        }

    def get_trainer(self) -> ITrainer:
        """
        Returns the agent's behavior.
        """
        return self._trainer

    def get_path(self) -> str:
        """
        Returns the agent's model save path.
        """
        return self._save_path

    def set_do_logging(self, do_logging: bool):
        """
        Sets the logging mode.

        Args:
            do_logging: True to enable logging, otherwise False.
        """
        self._logging = do_logging

    def execute(self, feature, scenario):
        """
        Executes the training or testing process based on the current mode (TRAIN_MODE or TEST_MODE).

        Args:
            feature: The specific feature to use.
            scenario: The specific scenario to use.
        """
        if not self._trainer:
            raise Exception("Trainer not set")

        _path = self._save_path
        if not _path:
            _path = model_path(feature, scenario)
        else:
            _path = _path + "/" + feature.name + "/" + scenario.name

        logging_path = None
        if self._logging:
            logging_path = _path + "/logs"

        try:
            if self._mode == TRAIN_MODE:
                self._trainer.train(_path, self._eval_freq, logs_path=logging_path, logs_name=scenario.name,
                                    previous_model_path=self._previous_model_path)
            elif self._mode == TEST_MODE:
                self._trainer.load(_path + "/best_model")
                self._trainer.play()
        except KeyboardInterrupt or SystemExit as e:
            logging.info("Training interrupted by user.")
            raise e
        except Exception as e:
            logging.error(f"Error occurs while {self._mode}: ", e)
            raise e
        finally:
            self._trainer.free()
            self.test_runner().finished()


    def predict(self, observation):
        """
        Predicts the next action based on the given observation.

        Args:
            observation: The observation to predict the next action.

        Returns:
            The predicted action.
        """
        return self._trainer.predict(observation)

    @abstractmethod
    def observation(self) -> OBST:
        """
        Abstract method to obtain observations from the environment.

        Returns:
            OBST: The observations from the environment.
        """
        raise NotImplementedError

    @abstractmethod
    def reward(self) -> float:
        """
        Abstract method to obtain the agent's current reward.

        Returns:
            float: The current reward.
        """
        raise NotImplementedError

    @abstractmethod
    def terminated(self) -> bool:
        """
        Abstract method to check if the episode is terminated.

        Returns:
            bool: True if the episode is terminated, otherwise False.
        """
        raise NotImplementedError

    @abstractmethod
    def actions(self, raws_actions) -> List[str]:
        """
        Abstract method to obtain the list of possible actions for the agent.

        Args:
            raws_actions: The raw actions to process.

        Returns:
            List[str]: The list of possible actions.
        """
        raise NotImplementedError



from gymnasium.envs.registration import register
import gymnasium

if 'xumes-v0' not in gymnasium.registry:
    register(
        id="xumes-v0",
        entry_point="gym_env.gym_adapter_env.gym_adapter:GymAdapter",
    )



from typing import TypeVar

ST = TypeVar("ST")


class IStateEntity:
    """
    The `IStateEntity` interface defines the methods for managing the state of an entity in the game.

    Methods:
        update: Updates the entity's attributes from a state.
        build: Builds an `IStateEntity` object from a state.
    """
    def update_state(self, state: ST):
        """
        Updates his attributes from a state.
        :param state: representation of the updates.
        """
        raise NotImplementedError

    @staticmethod
    def build(state: ST):
        """
        Build a IStateEntity from a state (initiate attributes).
        :param state: representation of the first state of the game object.
        """
        raise NotImplementedError



from xumes.assertions.assertion_bucket import AssertionBucket
from xumes.entity.entity_manager import EntityRegistry
from xumes.test_automation.driver import Driver


class TestContext:

    def __init__(self, entity_registry: EntityRegistry, driver: Driver, assertion_bucket: AssertionBucket):
        self._entity_registry = entity_registry
        self.driver = driver
        self._assertion_bucket = assertion_bucket

    def __getattr__(self, item):
        return self._entity_registry.get(item)

    def assert_true(self, condition: bool) -> None:
        self._assertion_bucket.assert_true(data=condition)

    def assert_false(self, condition: bool) -> None:
        self._assertion_bucket.assert_false(data=condition)

    def assert_equal(self, actual, expected) -> None:
        self._assertion_bucket.assert_equal(data=actual, expected=expected)

    def assert_not_equal(self, actual, expected) -> None:
        self._assertion_bucket.assert_not_equal(data=actual, expected=expected)

    def assert_greater(self, actual, expected) -> None:
        self._assertion_bucket.assert_greater_than(data=actual, expected=expected)

    def assert_greater_equal(self, actual, expected) -> None:
        self._assertion_bucket.assert_greater_than_or_equal(data=actual, expected=expected)

    def assert_less(self, actual, expected) -> None:
        self._assertion_bucket.assert_less_than(data=actual, expected=expected)

    def assert_less_equal(self, actual, expected) -> None:
        self._assertion_bucket.assert_less_than_or_equal(data=actual, expected=expected)

    def assert_between(self, actual, expected_min, expected_max) -> None:
        self._assertion_bucket.assert_between(data=actual, expected_min=expected_min, expected_max=expected_max)

    def assert_not_between(self, actual, expected_min, expected_max) -> None:
        self._assertion_bucket.assert_not_between(data=actual, expected_min=expected_min,
                                                  expected_max=expected_max)



import logging
from abc import ABC
from typing import final, TypeVar, Optional

import gymnasium as gym
import stable_baselines3
from stable_baselines3.common.callbacks import EvalCallback
from stable_baselines3.common.monitor import Monitor

# noinspection PyUnresolvedReferences
import xumes
from xumes.core.errors.running_ends_error import RunningEndsError
from xumes.modules.reinforcement_learning.agent_trainer import AgentTrainer

OBST = TypeVar("OBST")


class StableBaselinesTrainer(AgentTrainer, ABC):

    def __init__(self,
                 agent,
                 observation_space=None,
                 action_space=None,
                 max_episode_length: int = 1000,
                 total_timesteps: int = 1000000,
                 algorithm_type: str = "MultiInputPolicy",
                 algorithm=stable_baselines3.PPO,
                 ):
        super().__init__(agent)
        if observation_space is not None and action_space is not None:
            self.env = Monitor(gym.make(
                id="xumes-v0",
                max_episode_steps=max_episode_length,
                training_service=self,
                observation_space=observation_space,
                action_space=action_space,
            ), filename=None, allow_early_resets=True)
        self.algorithm = algorithm
        self.algorithm_type = algorithm_type
        self.total_timesteps = total_timesteps

        self.model = None

        self.observation_space = observation_space
        self.action_space = action_space
        self.max_episode_length = max_episode_length

        self.made = False

    @final
    def make(self):
        if self.observation_space is None or self.action_space is None:
            raise Exception("Observation space and action space must be set before calling make")
        self.env = Monitor(gym.make(
            id="xumes-v0",
            max_episode_steps=self.max_episode_length,
            training_service=self,
            observation_space=self.observation_space,
            action_space=self.action_space,
        ), filename=None, allow_early_resets=True)
        self.made = True

    def train(self, save_path: str = None, eval_freq: int = 1000, logs_path: Optional[str] = None,
              logs_name: Optional[str] = None, previous_model_path: Optional[str] = None):

        if not self.made:
            self.make()

        eval_callback = None
        if save_path:
            eval_callback = EvalCallback(self.env, best_model_save_path=save_path,
                                         log_path=save_path, eval_freq=eval_freq,
                                         deterministic=True, render=False)

        if previous_model_path:
            self.model = self.algorithm(self.algorithm_type, self.env, verbose=1, tensorboard_log=logs_path).load(
                previous_model_path, env=self.env).learn(
                self.total_timesteps,
                callback=eval_callback,
                tb_log_name=logs_name,
            )
        else:
            self.model = self.algorithm(self.algorithm_type, self.env, verbose=1, tensorboard_log=logs_path).learn(
                self.total_timesteps,
                callback=eval_callback,
                tb_log_name=logs_name,
            )

        self.finished()

    def save(self, path: str):
        self.model.save(path)

    def free(self):
        if self.env is not None:
            self.env.close()
        self.model = None

    def load(self, path: str):
        if not self.made:
            self.make()

        self.model = self.algorithm(self.algorithm_type, self.env, verbose=1).load(path, env=self.env)

    def play(self, timesteps: Optional[int] = None):

        if not self.made:
            self.make()

        obs, _ = self.env.reset()

        running = True

        def step():
            nonlocal obs
            nonlocal running

            action, _states = self.model.predict(obs, deterministic=True)
            try:
                obs, reward, terminated, done, info = self.env.step(action)
                if done or terminated:
                    self.env.reset()
            except RunningEndsError:
                logging.info(f"Received stop signal. Closing environment.")
                self.env.close()
                running = False
            except Exception as e:
                logging.error(f"Error in step: {e}")
                self.env.close()
                running = False

        if not timesteps:
            while running:
                step()
        else:
            for _ in range(timesteps):
                step()

    def finished(self):
        self.agent.test_runner().finished()

    def predict(self, obs: OBST):
        """
        Predicts the action to take given the observation.
        Method used only when using the model in non-training model, inside a scripted bot.
        """
        return self.model.predict(obs)



import logging
import textwrap
import time
from abc import abstractmethod

import multiprocess
from typing import List, Dict, Tuple

from tabulate import tabulate

from xumes.core.colors import bcolors
from xumes.core.modes import TEST_MODE, RENDER_MODE
from xumes.assertions.assertion_bucket import AssertionReport
from xumes.test_automation.feature_strategy import FeatureStrategy, Scenario
from xumes.communication.i_com_game_engine import IComGameEngine
from xumes.test_automation.registries import config_registry, given_registry, \
    when_registry, then_registry

from xumes.communication.implementations.socket_impl.com_game_instance_socket import ComGameInstanceSocket

from xumes.test_automation.game_instance_service import GameInstanceService
from xumes.test_automation.test_runner import TestRunner


class ScenarioData:

    def __init__(self, test_runner: GameInstanceService = None, process: multiprocess.Process = None, ip: str = None,
                 port: int = None, pid: int = None):
        self.test_runner = test_runner
        self.process = process
        self.ip = ip
        self.port = port
        self.pid = pid


class TestManager:
    """
    A class that manages the execution of tests in a game environment.

    The TestManager class is responsible for loading and running tests in a game environment. It provides functionality
    for creating game services, running tests, and managing communication with the training manager.

    Args:
        communication_service (ICommunicationServiceTestManager): An implementation of the
            ICommunicationServiceTestManager interface for communication with the training manager.
        mode (str, optional): The mode of the test execution. Can be 'test', 'render', or 'train'.
            Defaults to 'test'.
        timesteps (int, optional): The maximum number of steps to run in a test. Defaults to None.
        iterations (int, optional): The maximum number of iterations to run a test. Defaults to None.

    Methods:
        get_port(feature: str, scenario: str) -> int:
            Retrieves the port number for a given feature and scenario.
        add_test_runner_data(steps: str, ip: str, port: int) -> None:
            Adds game service data to the list of game services data.
        create_test_runner(steps: str, ip: str, port: int) -> GameService:
            Creates a game service instance with the specified steps, IP, and port.
        _build_test_runner(test_automation, ip, port) -> GameService:
            Abstract method to build a game service instance. Must be implemented by subclasses.
        test_all() -> None:
            Runs all the tests in the game environment.
        delete_test_runners() -> None:
            Deletes all game service instances.
        run_test(steps: str, active_processes) -> None:
            Runs a test with the given steps.
        run_test_render(steps: str, active_processes) -> None:
            Runs a test in render mode with the given steps.
    """

    def __init__(self, comm_service: IComGameEngine,
                 feature_strategy: FeatureStrategy,
                 mode: str = TEST_MODE, timesteps=None, iterations=None, do_logs: bool = False,
                 logging_level=logging.NOTSET, fps_limit=-1, render=False):
        self._comm_service: IComGameEngine = comm_service
        self._scenario_datas: Dict[Scenario, ScenarioData] = {}
        self._mode = mode
        self._timesteps = timesteps
        self._iterations = iterations
        self._feature_strategy: FeatureStrategy = feature_strategy
        self._assertion_queue = multiprocess.Queue()
        self.do_logs = do_logs
        self._logging_level = logging_level
        self._delta_time = 0
        self._fps_limit = fps_limit
        self._render = render
        self._features = []

    def add_test_runner_data(self, scenario: Scenario, ip: str, port: int):
        # Add a game service data to the list of game services data
        self._scenario_datas[scenario] = ScenarioData(ip=ip, port=port)

    def _init_logging(self):
        logging.basicConfig(format='%(levelname)s:%(message)s', level=self._logging_level)

    def test_all(self, path) -> None:
        try:
            test_time_start = time.time()

            # Retrieve features and scenarios
            self._feature_strategy.retrieve_feature(path)
            self._features = self._feature_strategy.features

            # Check if all tests are finished
            active_processes = multiprocess.Value('i', 0)

            # For all scenarios in each feature, we run the test
            for feature in self._features:
                logging.info(f"{bcolors.OKBLUE + bcolors.BOLD}Feature: {feature.name}{bcolors.ENDC}")

                # Run the feature (exec all scenarios)
                self._run_feature(feature, active_processes)

                # Wait for all tests to be finished
                while active_processes.value > 0:
                    time.sleep(1)

                # Stop all scenarios
                feature_scenarios_data = []
                for scenario in feature.scenarios:
                    feature_scenarios_data.append(self._scenario_datas[scenario])
                    self._scenario_datas.pop(scenario)

                self._comm_service.stop_scenarios([scenario_data.pid for scenario_data in feature_scenarios_data])

                for process in feature.processes:
                    process.join()
                    process.terminate()

            test_time_end = time.time()
            self._delta_time = round(test_time_end - test_time_start, 3)

            if self._mode == TEST_MODE or self._mode == RENDER_MODE:
                self._assert()
            else:
                converted_time = time.strftime("%H:%M:%S", time.gmtime(self._delta_time))
                print(f"{bcolors.OKGREEN}Training finished in {converted_time}s.{bcolors.ENDC}")

        except KeyboardInterrupt as e:
            logging.info(f"{bcolors.FAIL}Xumes interrupted by user.{bcolors.ENDC}")
        except Exception as e:
            logging.error(f"{bcolors.FAIL}Error occurs while testing: {e}{bcolors.ENDC}")
        finally:
            # Close all game processes
            self._comm_service.stop_scenarios([scenario_data.pid for scenario_data in self._scenario_datas.values()])

            # Close all processes
            for feature in self._features:
                for process in feature.processes:
                    process.join()
                    process.terminate()

    def list_features(self, path):
        try:
            # Retrieve features and scenarios
            self._feature_strategy.retrieve_feature(path)
            self._features = self._feature_strategy.features
            width = 80

            # Create a list of rows for the table
            table_data = []
            for feature in self._features:
                header = f"{bcolors.OKBLUE + bcolors.BOLD}Feature: {feature.name}{bcolors.ENDC}"
                table_data.append([header, "", ""])
                for scenario in feature.scenarios:
                    tags = ', '.join(scenario.tags)
                    wrapped_tags = textwrap.fill(tags, width=50, subsequent_indent=' ' * 8)
                    table_data.append([f"  {scenario.name}", f"{scenario.steps}.py", wrapped_tags])

            # Define table headers
            headers = [bcolors.UNDERLINE + bcolors.BOLD + "Feature/Scenario" + bcolors.ENDC,
                       bcolors.UNDERLINE + bcolors.BOLD + "Steps file" + bcolors.ENDC,
                       bcolors.UNDERLINE + bcolors.BOLD + "Tags" + bcolors.ENDC]

            # Print the table
            print(tabulate(table_data, headers=headers, tablefmt="grid"))
        except Exception as e:
            logging.error(f"{bcolors.FAIL}Error occurs while listing features: {e}{bcolors.ENDC}")

    def _assert(self):
        # Make assertions

        results: List[AssertionReport] = []
        successes = 0
        tests_passed_names = ''
        error_logs = ''

        # Gather all assertion reports
        while not self._assertion_queue.empty():
            assertion_report = self._assertion_queue.get()
            if assertion_report is None:
                break
            results.append(assertion_report)
            if assertion_report.passed:
                successes += 1
                tests_passed_names += '    - ' + assertion_report.test_name + '\n'
            else:
                error_logs += assertion_report.error_logs

        # log results
        nb_test = len(results)
        converted_time = time.strftime("%H:%M:%S", time.gmtime(self._delta_time))
        header = f"{bcolors.BOLD}{bcolors.UNDERLINE}{'':15}TEST REPORT{'':15}{bcolors.ENDC}\n"
        details = f"{successes} tests passed on a total of {nb_test} in {converted_time}.\n"
        details += f"Tests passed:\n{tests_passed_names}\n" if successes > 0 else ""
        details += error_logs

        if successes < nb_test:
            print(f"{bcolors.FAIL}{header}{bcolors.ENDC}")
            print(f"{bcolors.FAIL}{details}{bcolors.ENDC}")
        else:
            print(f"{bcolors.OKGREEN}{header}{bcolors.ENDC}")
            print(f"{bcolors.OKGREEN}{details}{bcolors.ENDC}")

    def _run_feature(self, feature, active_processes):

        scenarios_methods: List[Dict] = []

        for scenario in feature.scenarios:
            self.add_test_runner_data(scenario=scenario, ip="127.0.0.1", port=0)

            registry_queue = multiprocess.Queue()
            registry_queue.put((config_registry, given_registry, when_registry, then_registry))

            test_runner = TestRunner(GameInstanceService(ComGameInstanceSocket(host="127.0.0.1")),
                                     self._timesteps, self._iterations, self._mode, scenario.feature.name,
                                     scenario.name,
                                     self._assertion_queue, scenario.steps, registry_queue)

            self._scenario_datas[scenario].test_runner = test_runner

            methods = test_runner.config()

            scenarios_methods.append({
                "name": scenario.name,
                "methods": methods,
                "fps_limit": self._fps_limit,
                "render": self._render
            })

        # Make call to api to start the game instances
        ports_pids: Dict[scenario, Tuple[int, int]] = self._comm_service.start_scenarios(scenarios_methods,
                                                                                         feature.scenarios,
                                                                                         self._fps_limit, self._render)

        for scenario in ports_pids:
            self._scenario_datas[scenario].port = ports_pids[scenario][0]
            self._scenario_datas[scenario].pid = ports_pids[scenario][1]
            self._scenario_datas[scenario].test_runner.run(ports_pids[scenario][0])

        self._run_scenarios(feature=feature, scenario_datas=self._scenario_datas, active_processes=active_processes)

    @abstractmethod
    def _run_scenarios(self, feature, scenario_datas, active_processes):
        raise NotImplementedError


class DummyTestManager(TestManager):
    def _run_scenarios(self, feature, scenario_datas, active_processes):
        pass



import os
import re
from typing import Dict, List

from gherkin.parser import Parser
from gherkin.token_scanner import TokenScanner

from xumes.core.registry import get_content_from_registry
from xumes.test_automation.feature_strategy import FeatureStrategy, Feature, Scenario


class GherkinFeatureStrategy(FeatureStrategy):
    """
    Gherkin language implementation of FeatureStrategy.
    Using the .feature files, we can get all features and scenarios.
    It also implements a way to filter the features and scenarios that we want to run.
    """
    def __init__(self, alpha: float = 0.001, features_names: List[str] = None, scenarios_names: List[str] = None,
                 tags: List[str] = None, steps_path: str = "./"):
        super().__init__(alpha, steps_path)
        self._selected_features: List[str] = features_names
        self._selected_scenarios: List[str] = scenarios_names
        self._selected_tags: List[str] = tags
        self._steps = []

    def retrieve_feature(self, path: str):
        """
        Get features and scenarios from every feature file ".feature" in the "./features" directory.
        """

        # Convert steps files to list of gherkin steps
        self._convert_steps_files_to_gherkin_steps()

        # Get all feature files
        feature_files = os.listdir(path)
        feature_files = [file for file in feature_files if file.endswith(".feature")]

        # If there are no feature files, raise an exception
        if len(feature_files) == 0:
            raise Exception("Feature files not found.")

        # Initialize parser
        parser = Parser()

        # Iterate over feature files
        for feature_file in feature_files:
            if self._selected_features is None or feature_file[:-8] in self._selected_features:
                # We open the feature file and parse it
                feature_path = os.path.join(path, feature_file)

                with open(feature_path, 'r') as file:
                    feature = parser.parse(TokenScanner(file.read()))['feature']

                    # Create a Feature object
                    feature_obj = Feature(name=feature_file[:-8])

                    # Iterate over every scenario in the feature file
                    for element in feature['children']:
                        if 'scenario' in element:
                            scenario = element['scenario']
                            if scenario['keyword'] == 'Scenario Outline':
                                # Handle Scenario Outline
                                self._process_scenario_outline(scenario, feature_obj)
                            elif scenario['keyword'] == 'Scenario':
                                # Handle regular Scenario
                                self._process_scenario(scenario, feature_obj)

                    # Append the feature to the list of features
                    self.features.append(feature_obj)

    def _process_scenario_outline(self, scenario_outline: Dict, feature_obj: Feature):
        """
        Process a Scenario Outline and generate scenarios for each example row.
        """
        examples = scenario_outline['examples']
        example_count = 0  # Initialize example counter for this scenario outline
        for example in examples:
            headers = [cell['value'] for cell in example['tableHeader']['cells']]
            for row in example['tableBody']:
                example_values = {headers[i]: cell['value'] for i, cell in enumerate(row['cells'])}
                example_description = ', '.join([f"{key}={value}" for key, value in example_values.items()])
                scenario_name = f"{scenario_outline['name']} [{example_count}] - ({example_description})"
                example_count += 1  # Increment example counter for this scenario outline
                scenario_copy = self._create_scenario_from_outline(scenario_outline, example_values, scenario_name)
                if self._selected_scenarios is None or scenario_copy['name'] in self._selected_scenarios:
                    self._process_scenario(scenario_copy, feature_obj, is_outline=True)

    def _create_scenario_from_outline(self, scenario_outline: Dict, example_values: Dict[str, str], scenario_name: str) -> Dict:
        """
        Create a scenario from a scenario outline by replacing placeholders with example values.
        """
        scenario_copy = {
            'name': scenario_name,
            'steps': [],
            'tags': scenario_outline.get('tags', [])
        }
        for step in scenario_outline['steps']:
            step_text = step['text']
            for placeholder, value in example_values.items():
                step_text = step_text.replace(f"<{placeholder}>", value)
            scenario_copy['steps'].append({
                'keyword': step['keyword'],
                'keywordType': step['keywordType'],
                'text': step_text
            })
        return scenario_copy

    def _process_scenario(self, scenario: Dict, feature_obj: Feature, is_outline: bool = False):
        """
        Process a single scenario and add it to the feature object.
        """
        # If the scenario has tags, we check if the scenario has one of the selected tags
        has_tag = False
        scenario_tags = []
        if 'tags' in scenario:
            for tag in scenario.get('tags', []):
                scenario_tags.append(tag['name'][1:])

        if self._selected_tags is None:
            has_tag = True
        else:
            for tag in scenario_tags:
                if tag in self._selected_tags:  # We remove the @ from the tag
                    has_tag = True
                    break
            else:
                has_tag = False

        # Check if the scenario name matches the selected scenarios
        if (self._selected_scenarios is None or scenario['name'] in self._selected_scenarios) and has_tag:
            # Find the steps file for the scenario using pattern matching
            steps_file, given_params, when_params, then_params = self._find_steps_file(scenario, feature_obj.name)

            # Fill parameters
            for i in range(len(given_params)):
                self.given.all[steps_file][i].add_params(scenario['name'], given_params[i])
            for i in range(len(when_params)):
                self.when.all[steps_file][i].add_params(scenario['name'], when_params[i])
            for i in range(len(then_params)):
                self.then.all[steps_file][i].add_params(scenario['name'], then_params[i])

            # Create a Scenario object
            scenario_obj = Scenario(scenario['name'], steps_file, feature_obj, scenario_tags)
            feature_obj.scenarios.append(scenario_obj)

    def _convert_steps_files_to_gherkin_steps(self):
        """
        Convert steps files to list of gherkin steps
        Just taking the content of the steps files and putting in the gherkin steps list
        """
        self._steps.clear()
        for name in self.given.all:
            try:
                given = self.given.all[name]
            except IndexError:
                raise Exception("Given steps not found.")
            try:
                when = self.when.all[name]
            except IndexError:
                raise Exception("When steps not found.")
            try:
                then = self.then.all[name]
            except IndexError:
                raise Exception("Then steps not found.")

            given_contents = get_content_from_registry(given)
            when_contents = get_content_from_registry(when)
            then_contents = get_content_from_registry(then)

            self._steps.append({
                "name": name,
                "given_contents": given_contents,
                "when_contents": when_contents,
                "then_contents": then_contents
            })

    def _find_steps_file(self, scenario: Dict, feature_name: str):
        """
        Find the steps file for the scenario using pattern matching.
        """
        # Get scenario name
        previous_type = None
        scenario_givens = []
        scenario_whens = []
        scenario_thens = []

        # Iterate over steps
        # And convert the steps to a list of gherkin steps
        # Check the gherkin-official package for more information
        for step in scenario['steps']:
            if step['keywordType'] == "Context" or (
                    step['keywordType'] == "Conjunction" and previous_type == "Context"):
                scenario_givens.append(step['text'])
                previous_type = "Context"
            elif step['keywordType'] == "Action" or (
                    step['keywordType'] == "Conjunction" and previous_type == "Action"):
                scenario_whens.append(step['text'])
                previous_type = "Action"
            elif step['keywordType'] == "Outcome" or (
                    step['keywordType'] == "Conjunction" and previous_type == "Outcome"):
                scenario_thens.append(step['text'])
                previous_type = "Outcome"
            else:
                raise Exception("Keyword type not found.")

        # Iterate over steps files
        for steps_file in self._steps:
            # Get steps file name
            steps_file_name = steps_file['name']
            steps_file_givens = steps_file['given_contents']
            steps_file_whens = steps_file['when_contents']
            steps_file_thens = steps_file['then_contents']

            # Check if the scenario contents corresponds to the steps file contents
            given_parameters = self._steps_matching(steps_file_givens, scenario_givens)
            when_parameters = self._steps_matching(steps_file_whens, scenario_whens)
            then_parameters = self._steps_matching(steps_file_thens, scenario_thens)

            if given_parameters != False and when_parameters != False and then_parameters != False:
                return steps_file_name, given_parameters, when_parameters, then_parameters
        # If no steps file was found, raise an exception
        raise Exception(f"Steps file for scenario: {feature_name}/{scenario['name']} not found.")

    @staticmethod
    def _steps_matching(step_files_steps, scenario_steps):
        # Check if the step files steps corresponds to the scenario steps
        # And return parameters if it does
        if len(step_files_steps) == len(scenario_steps):
            tmp = []
            for i in range(len(step_files_steps)):
                param = GherkinFeatureStrategy._pattern_matching(step_files_steps[i], scenario_steps[i])
                # noinspection PySimplifyBooleanCheck
                if param != False:
                    tmp.append(param)
                else:
                    tmp = False
                    break
            return tmp
        else:
            return False

    @staticmethod
    def _pattern_matching(step: str, scenario: str):
        # Check if the step str corresponds to the scenario str
        # And return parameters if it does

        # Pattern to match parameters with type
        param_pattern = re.compile(r'\{(\w+):(\w+)\}')

        # Split step and scenario
        step_parts = step.split(" ")
        scenario_parts = scenario.split(" ")

        # If the length of the step is different from the length of the scenario, return False
        if len(step_parts) != len(scenario_parts):
            return False

        # Iterate over step and scenario
        parameters = {}
        for i in range(len(step_parts)):
            match = param_pattern.match(step_parts[i])
            if match:
                param_name, param_type = match.groups()
                param_value = scenario_parts[i]

                # Convert the parameter value to the appropriate type
                if param_type == 'd':
                    param_value = int(param_value)
                elif param_type == 'f':
                    param_value = float(param_value)
                elif param_type == 's':
                    param_value = str(param_value)
                else:
                    return False

                parameters[param_name] = param_value
            elif step_parts[i] != scenario_parts[i]:
                return False

        # If the step and scenario are equal, return the parameters those can be empty
        return parameters



import multiprocess

from xumes.test_automation.test_manager import TestManager
from xumes.modules.reinforcement_learning.agent import Agent


class RLTestManager(TestManager):

    def _run_scenarios(self, feature, scenario_datas, active_processes):
        for scenario in scenario_datas:
            feature = scenario.feature
            test_runner = scenario_datas[scenario].test_runner

            def run(nb_processes, t_runner, logging):
                try:
                    when_result = t_runner.when()
                    if len(when_result) > 1:
                        raise Exception("Only one when step is allowed")

                    behavior: Agent = when_result[next(iter(when_result))]
                    behavior.set_mode(self._mode)
                    behavior.set_test_runner(t_runner)
                    behavior.set_do_logging(logging)
                    behavior.execute(scenario.feature, scenario)
                except Exception as e:
                    raise e
                finally:
                    nb_processes.value -= 1
                    exit(0)

            process = multiprocess.Process(target=run, args=(active_processes, test_runner, self.do_logs))
            process.start()
            feature.processes.append(process)
            active_processes.value += 1



from xumes.assertions.assertion import IAssertionStrategy, AssertionEqualStatistical, AssertionBetweenStatistical, \
    AssertionLessThanStatistical, AssertionLessThanOrEqualStatistical, AssertionGreaterThanStatistical, \
    AssertionGreaterThanOrEqualStatistical, AssertionGreaterThanOrEqual, AssertionGreaterThan, AssertionLessThanOrEqual, \
    AssertionLessThan, AssertionBetween, AssertionEqual


class IAssertionFactory:

    def assertion_equal(self, value) -> IAssertionStrategy:
        raise NotImplementedError

    def assertion_between(self, min_value, max_value) -> IAssertionStrategy:
        raise NotImplementedError

    def assertion_less_than(self, value) -> IAssertionStrategy:
        raise NotImplementedError

    def assertion_less_than_or_equal(self, value) -> IAssertionStrategy:
        raise NotImplementedError

    def assertion_greater_than(self, value) -> IAssertionStrategy:
        raise NotImplementedError

    def assertion_greater_than_or_equal(self, value) -> IAssertionStrategy:
        raise NotImplementedError


class AssertionFactoryStatistical(IAssertionFactory):

    def __init__(self, alpha=0.001):
        self._alpha = alpha

    def assertion_equal(self, value) -> IAssertionStrategy:
        return AssertionEqualStatistical(value, alpha=self._alpha)

    def assertion_between(self, min_value, max_value) -> IAssertionStrategy:
        return AssertionBetweenStatistical(min_value, max_value, alpha=self._alpha)

    def assertion_less_than(self, value) -> IAssertionStrategy:
        return AssertionLessThanStatistical(value, alpha=self._alpha)

    def assertion_less_than_or_equal(self, value) -> IAssertionStrategy:
        return AssertionLessThanOrEqualStatistical(value, alpha=self._alpha)

    def assertion_greater_than(self, value) -> IAssertionStrategy:
        return AssertionGreaterThanStatistical(value, alpha=self._alpha)

    def assertion_greater_than_or_equal(self, value) -> IAssertionStrategy:
        return AssertionGreaterThanOrEqualStatistical(value, alpha=self._alpha)


class AssertionFactory(IAssertionFactory):

    def assertion_equal(self, value) -> IAssertionStrategy:
        return AssertionEqual(value)

    def assertion_between(self, min_value, max_value) -> IAssertionStrategy:
        return AssertionBetween(min_value, max_value)

    def assertion_less_than(self, value) -> IAssertionStrategy:
        return AssertionLessThan(value)

    def assertion_less_than_or_equal(self, value) -> IAssertionStrategy:
        return AssertionLessThanOrEqual(value)

    def assertion_greater_than(self, value) -> IAssertionStrategy:
        return AssertionGreaterThan(value)

    def assertion_greater_than_or_equal(self, value) -> IAssertionStrategy:
        return AssertionGreaterThanOrEqual(value)



import os

from xumes.test_automation.feature_strategy import FeatureStrategy, Feature, Scenario


class BasicFeatureStrategy(FeatureStrategy):

    def retrieve_feature(self):
        """
        Get features and scenarios from a feature file ".feature" in the test directory.
        The feature file must be in the following format:
        Feature: <feature_name>
            Scenario: <scenario_name>
            Scenario: <scenario_name>
        Feature: <feature_name>
            Scenario: <scenario_name>
            Scenario: <scenario_name>
        """
        # Clear features
        self.features.clear()

        # Check if feature file exists
        if not os.path.exists(".features"):
            raise Exception("Feature file not found.")

        # Open feature file
        with open(".features", 'r') as feature_file:
            # Read all lines
            lines = feature_file.readlines()
            # Initialize feature
            feature = None
            # Iterate over lines
            for line in lines:
                # Remove spaces and \n
                line = line.strip()

                if line.startswith("Feature:"):

                    # Save the previous feature
                    if feature is not None:
                        self.features.append(feature)

                    # If line starts with "Feature:" then create a new feature
                    feature_name = line.replace("Feature:", "").strip()
                    feature = Feature(name=feature_name)
                elif line.startswith("Scenario:"):
                    # If line starts with "Scenario:" then create a new scenario
                    scenario_name = line.replace("Scenario:", "").strip()
                    scenario = Scenario(name=scenario_name, steps=scenario_name, feature=feature)
                    feature.scenarios.append(scenario)

            if feature is not None:
                self.features.append(feature)



from abc import ABC
from typing import final, TypeVar

from xumes.modules.reinforcement_learning.i_trainer import ITrainer

OBST = TypeVar("OBST")


class AgentTrainer(ITrainer, ABC):

    def __init__(self, agent):
        self.agent = agent

    @final
    def get_reward(self):
        return self.agent.reward()

    @final
    def do_reset(self):
        self.agent.test_runner().reset()

    @final
    def episode_finished(self):
        return self.agent.test_runner().episode_finished()

    @final
    def get_terminated(self):
        return (self.agent.terminated() or self.agent.test_runner().game_state == "reset"
                or self.agent.test_runner().game_state == "random_reset")

    @final
    def get_obs(self) -> OBST:
        self.agent.test_runner().retrieve_state()
        return self.agent.observation()

    @final
    def push_actions_and_get_obs(self, actions) -> OBST:
        self.agent.test_runner().push_action_and_get_state(self.agent.actions(actions))
        return self.agent.observation()



from gymnasium.envs.registration import register

register(
    id="xumes-v0",
    entry_point="xumes.modules.reinforcement_learning.gym_impl.gym_envs.gym_env.gym_adapter_env.gym_adapter:GymAdapter",
)



import json
from socket import socket, AF_INET, SOCK_STREAM
from time import sleep
from typing import Dict, Any

from xumes.core.utils import parse_json_with_eval
from xumes.communication.i_com_game_instance import IComGameInstance


class ComGameInstanceSocket(IComGameInstance):

    def __init__(self, host: str):
        self.host = host
        self.socket = None
        self.addr = None
        self.is_running = False

    def push_dict(self, dictionary) -> None:

        if self.is_running:
            self.socket.sendall(json.dumps(dictionary).encode())

    def get_dict(self) -> Dict[str, Any]:
        data = {}
        if self.is_running:
            data = self.socket.recv(1024)
            data = data.decode()
            data = json.loads(data)
            data = parse_json_with_eval(data)
        return data

    def get_int(self) -> int:
        data = 0
        if self.is_running:
            data = self.socket.recv(1024)
            data = eval(data)
        return data

    def init_socket(self, port) -> None:
        self.socket = socket(AF_INET, SOCK_STREAM)
        self.socket.settimeout(1000)
        while True:
            try:
                self.socket.connect((self.host, port))
                break
            except Exception as e:
                sleep(0.001)
        self.is_running = True

    def stop_socket(self) -> None:
        """
        Used to stop the communication service.
        """
        self.is_running = False
        self.socket.close()



import time

import multiprocess
from typing import List, Dict

from xumes.assertions.assertion import IAssertionStrategy
from xumes.assertions.assertion_factory import AssertionFactory
from xumes.assertions.assertion_result import AssertionResult


class AssertionReport:

    def __init__(self, passed: bool, error_logs: str, test_name: str):
        self.passed = passed
        self.error_logs = error_logs
        self.test_name = test_name

    def __getstate__(self):
        return self.passed, self.error_logs, self.test_name

    def __setstate__(self, state):
        self.passed, self.error_logs, self.test_name = state


class AssertionBucketState:

    def __init__(self, assertion_bucket):
        self._assertion_bucket = assertion_bucket

    def action(self, *args, **kwargs):
        raise NotImplementedError


class CollectState(AssertionBucketState):

    def action(self, *args, **kwargs):
        if 'data' in kwargs:
            self._assertion_bucket.do_collect(kwargs['data'])
        else:
            raise ValueError("Missing data argument")


class AssertState(AssertionBucketState):

    def action(self, *args, **kwargs):
        if 'expected' in kwargs and 'assertion_strategy' in kwargs and 'opposite' in kwargs:
            self._assertion_bucket.do_assert(kwargs['expected'], kwargs['assertion_strategy'], kwargs['opposite'])
        elif 'expected' in kwargs and 'assertion_strategy' in kwargs:
            self._assertion_bucket.do_assert(kwargs['expected'], kwargs['assertion_strategy'])
        else:
            raise ValueError("Missing expected or assertion_strategy argument")


class AssertionBucket:
    """
    A class that holds a list of lists of values to assert.
    Each list of values is a test case.
    We then iterate over each test case and assert the values.
    """
    ASSERT_MODE = "assert"
    COLLECT_MODE = "collect"

    def __init__(self, test_name, queue: multiprocess.Queue, assertion_factory=AssertionFactory()):
        super().__init__()
        self._data = []
        self._results: List[AssertionResult] = []
        self._iterator = 0
        self._test_name = test_name
        self._collect_state = CollectState(self)
        self._assert_state = AssertState(self)
        self._state = self._collect_state
        self._queue = queue
        self._passed = True
        self._assertion_factory: AssertionFactory = assertion_factory

    def reset_iterator(self):
        self._iterator = 0

    def collect_mode(self):
        self._state = self._collect_state

    def assertion_mode(self):
        self._state = self._assert_state

    def _collect_or_assert(self, data, expected, assertion_strategy: IAssertionStrategy, opposite=False):
        self._state.action(data=data, expected=expected, assertion_strategy=assertion_strategy, opposite=opposite)

    def assert_from_dict(self, assertion_dicts: List[Dict]):
        for assertion_dict in assertion_dicts:
            if assertion_dict['type'] == 'assert_equal':
                self.assert_equal(data=assertion_dict['actual'], expected=assertion_dict['expected'])
        self.reset_iterator()

    def assert_true(self, data):
        self._state.action(data=data, expected=True, assertion_strategy=self._assertion_factory.assertion_equal(True),
                           opposite=False)

    def assert_false(self, data):
        self._state.action(data=data, expected=False, assertion_strategy=self._assertion_factory.assertion_equal(False),
                           opposite=False)

    def assert_equal(self, data, expected):
        self._state.action(data=data, expected=expected,
                           assertion_strategy=self._assertion_factory.assertion_equal(expected))

    def assert_not_equal(self, data, expected):
        self._state.action(data=data, expected=expected,
                           assertion_strategy=self._assertion_factory.assertion_equal(expected), opposite=True)

    def assert_greater_than(self, data, expected):
        self._state.action(data=data, expected=expected,
                           assertion_strategy=self._assertion_factory.assertion_greater_than(expected))

    def assert_greater_than_or_equal(self, data, expected):
        self._state.action(data=data, expected=expected,
                           assertion_strategy=self._assertion_factory.assertion_greater_than_or_equal(expected))

    def assert_less_than(self, data, expected):
        self._state.action(data=data, expected=expected,
                           assertion_strategy=self._assertion_factory.assertion_less_than(expected))

    def assert_less_than_or_equal(self, data, expected):
        self._state.action(data=data, expected=expected,
                           assertion_strategy=self._assertion_factory.assertion_less_than_or_equal(expected))

    def assert_between(self, data, expected_min, expected_max):
        self._state.action(data, expected=(expected_min, expected_max),
                           assertion_strategy=self._assertion_factory.assertion_between(expected_min,
                                                                                        expected_max))

    def assert_not_between(self, data, expected_min, expected_max):
        self._state.action(data=data, expected=(expected_min, expected_max),
                           assertion_strategy=self._assertion_factory.assertion_between(expected_min,
                                                                                        expected_max),
                           opposite=True)

    def do_collect(self, other):
        if self._iterator < len(self._data):
            self._data[self._iterator].append(other)
        else:
            self._data.append([other])
        self._iterator += 1

    def do_assert(self, expected, assertion_strategy: IAssertionStrategy, opposite=False):
        # Get the actual value and assert it
        actual = self._data[self._iterator]
        r = assertion_strategy.test(actual)
        if opposite:  # If we want to assert the opposite
            r = not r
        if not r:
            self._passed = False
        self._results.append(AssertionResult(
            fail_message=f"Test {self._test_name} FAILED on {self._iterator + 1}th assertion",
            passed=r,
            actual=actual,
            expected=expected
        ))
        self._iterator += 1

    def send_results(self):
        error_logs = ""
        for assertion_result in self._results:
            if not assertion_result.passed:
                error_logs += f"\n{assertion_result.fail_message:50}\n" \
                              f"{'Actual':10}: {assertion_result.actual} \n" \
                              f"{'Expected':10}: {assertion_result.expected}\n"

        assertion_report = AssertionReport(passed=self._passed,
                                           error_logs=error_logs,
                                           test_name=self._test_name
                                           )

        self._queue.put(assertion_report)
        time.sleep(0.5)

    def clear(self):
        self._data.clear()
        self._results.clear()
        self._iterator = 0



import logging
from typing import Dict, List

import requests
from xumes.communication.i_com_game_engine import IComGameEngine


class ComGameEngineRest(IComGameEngine):

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.session = requests.Session()
        self.ports = []

    def start_scenario(self, name: str, methods: List[Dict], fps_limit: int, render: bool) -> int:
        port = 0
        body = {
            "methods": methods,
            "fps_limit": fps_limit,
            "render": render
        }
        http_response = self.session.post(f"http://{self.host}:{self.port}/start_scenario/", json=body)
        if http_response.status_code == 200:
            port = http_response.json()
            logging.log(logging.INFO, f"Scenario {name} started on port {port}.")
        return port

    def start_scenarios(self, scenarios_methods, scenarios, fps_limit: int, render: bool) -> Dict[str, int]:
        ports_pids = {}
        http_response = requests.post(f"http://{self.host}:{self.port}/start_scenarios/", json=scenarios_methods)
        if http_response.status_code == 200:
            ports_response = http_response.json()
            for scenario_method in scenarios_methods:
                scenario_name = scenario_method["name"]
                port_pid = ports_response[scenario_name]
                scenario = next(scenario for scenario in scenarios if scenario.name == scenario_name) # TODO change this to a dict
                ports_pids[scenario] = port_pid
                logging.log(logging.INFO, f"Scenario {scenario_name} started on port {port_pid[0]} with pid {port_pid[1]}.")

        return ports_pids

    def stop_scenario(self, port: int) -> None:
        http_response = self.session.post(f"http://{self.host}:{self.port}/stop_scenario/", json=port)
        if http_response.status_code == 200:
            logging.log(logging.INFO, f"Scenario on port {port} stopped.")

    def stop_scenarios(self, pids: List[int]) -> None:
        http_response = requests.post(f"http://{self.host}:{self.port}/stop_scenarios/", json=pids)
        if http_response.status_code == 200:
            logging.log(logging.INFO, "Scenarios stopped.")




class Driver:
    """
    A class that translates a function call into a JSON object containing the function name and parameters,
    and stores these calls in a dictionary under the key 'methods'.

    Attributes:
        _calls (list): A list to store the details of function calls.

    Methods:
        __getattr__(name: str) -> Callable:
            Returns a callable that captures the function name and parameters.
        __call__() -> list:
            Returns a list of stored function calls and clears the calls list.
    """

    def __init__(self):
        self._calls = []

    def __call__(self):
        """
        Returns a list containing the stored function calls and clears the calls list.

        Returns:
            list: A list containing the stored function calls.
        """
        call_details_list = self._calls
        self._calls = []
        return call_details_list

    def __getattr__(self, name: str):
        """
        Returns a callable that captures the function name and parameters.

        Args:
            name (str): The name of the function being called.

        Returns:
            Callable: A function that takes parameters and stores the call details.
        """

        def wrapper(*args, **kwargs):
            call_details = {
                "function_name": name,
                "parameters": {
                    "args": list(args),
                    "kwargs": kwargs
                }
            }
            self._calls.append(call_details)

        return wrapper



from typing import Tuple

from xumes.entity.game_element_state import GameElementState
from xumes.entity.i_game_element_state_converter import IGameElementStateConverter


class JsonGameElementStateConverter(IGameElementStateConverter):

    def convert(self, state_wrapper: Tuple) -> GameElementState:
        name, content = state_wrapper
        if isinstance(content, dict):
            if "__type__" in content:
                element_type = content["__type__"]
                state = {}
                for k in content:
                    state[k] = content[k]

                return GameElementState(name=name, element_type=element_type, state_obj=state)

        return GameElementState(name=name, element_type="unknown", state_obj=content)




import logging
import platform
from multiprocessing import set_start_method
from typing import List

import click
import re

from xumes.core.modes import TRAIN_MODE, TEST_MODE
from xumes.modules.human.human_playing_test_manager import HumanPlayingTestManager
from xumes.test_automation.implementations.features_impl.gherkin_feature_strategy import GherkinFeatureStrategy
from xumes.communication.implementations.rest_impl.com_game_engine_rest import ComGameEngineRest
from xumes.modules.reinforcement_learning.rl_test_manager import RLTestManager
from xumes.modules.reinforcement_learning.vec_rl_test_manager import VecRLTestManager
from xumes.test_automation.test_manager import DummyTestManager


@click.group()
def cli():
    pass


def get_debug_level(debug, info):
    if debug:
        return logging.DEBUG
    if info:
        return logging.INFO

    return logging.CRITICAL


def split_names(names: str) -> List[str]:
    """
    Two cases:
        - names='a,b,c' -> ["a", "b", "c"]
        - names='"a, a","b, b","c, c"' -> ["a, a", "b, b", "c, c"]
    Args:
        names (str): A string of names separated by commas. If a name contains a comma, it should be enclosed in quotes.

    Returns:
        List[str]: The list of names
    """
    pattern = r'[^,\s][^,]*[^,\s]*|"[^"]*"'
    return [name.strip('"') for name in re.findall(pattern, names)]


@cli.command()
@click.option("--vectorize", "-v", is_flag=True, help="Vectorize the training.")
@click.option("--iterations", "-i", default=None, help="Number of iterations to test the game.")
@click.option("--features", "-f", default=None, help="List of features to test.")
@click.option("--scenarios", "-s", default=None, help="List of scenarios to test.")
@click.option("--tags", default=None, help="Tags of the features to test.")
@click.option("--log", is_flag=True, help="Log the game.")
@click.option("--debug", is_flag=True, help="Debug debug level.")
@click.option("--info", is_flag=True, help="Info debug level.")
@click.option("--features_path", default=None, type=click.Path(), help="Path of the ./features folder.")
@click.option("--steps_path", default=None, type=click.Path(), help="Path of the ./steps folder.")
@click.option("--alpha", "-a", default=0.001, help="Alpha of the training.")
@click.option("--fps", default=-1, help="Limit of FPS.")
@click.option("--headless", is_flag=False, help="Run the game in headless mode.")
@click.option("-h", default=-1, help="Headless and fix FPS.")
def test(debug, features_path, steps_path, iterations, info, log, alpha,
         features,
         scenarios,
         tags,
         vectorize,
         fps, headless, h):
    if iterations is not None:
        iterations = int(iterations)
    else:
        raise Exception("Number of iterations must be set")

    # change start method to fork to avoid errors with multiprocessing
    # Windows does not support the fork start method
    if platform.system() != "Windows":
        set_start_method('fork')

    logging_level = get_debug_level(debug, info)
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging_level)

    mode = TRAIN_MODE

    if vectorize:
        test_manager_class = VecRLTestManager
    else:
        test_manager_class = RLTestManager

    if h > 0:
        headless = True
        fps = h

    render = not headless

    test_manager = test_manager_class(
        feature_strategy=GherkinFeatureStrategy(steps_path=steps_path,
                                                features_names=features,
                                                scenarios_names=scenarios,
                                                tags=tags),
        mode=TEST_MODE,
        logging_level=logging_level,
        iterations=iterations,
        render=render,
        fps_limit=fps,
        comm_service=ComGameEngineRest(host="127.0.0.1", port=8080)
    )

    test_manager.test_all(features_path)


@cli.command()
@click.option("--tensorboard", "-tb", is_flag=True, help="Save logs to _logs folder to be use with the tensorboard.")
@click.option("--vectorize", "-v", is_flag=True, help="Vectorize the training.")
@click.option("--debug", is_flag=True, help="Debug debug level.")
@click.option("--info", is_flag=True, help="Info debug level.")
@click.option("--features_path", default=None, type=click.Path(), help="Path of the ./features folder.")
@click.option("--steps_path", default=None, type=click.Path(), help="Path of the ./steps folder.")
@click.option("--model", default=None, type=click.Path(),
              help="Path of the model to load if you want to use a base model for your training.")
@click.option("--features", "-f", default=None, help="List of features to test.")
@click.option("--scenarios", "-s", default=None, help="List of scenarios to test.")
@click.option("--tags", default=None, help="Tags of the features to test.")
@click.option("--fps", default=-1, help="Limit of FPS.")
@click.option("--headless", is_flag=False, help="Run the game in headless mode.")
@click.option("-h", default=-1, help="Headless and fix FPS.")
def train(debug, steps_path, info, tensorboard, model, features_path, vectorize, features, scenarios, tags, fps,
          headless, h):
    # change start method to fork to avoid errors with multiprocessing
    # Windows does not support the fork start method
    if platform.system() != "Windows":
        set_start_method('fork')

    logging_level = get_debug_level(debug, info)
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging_level)

    if vectorize:
        test_manager_class = VecRLTestManager
    else:
        test_manager_class = RLTestManager

    if h > 0:
        headless = True
        fps = h

    render = not headless

    test_manager = test_manager_class(
        feature_strategy=GherkinFeatureStrategy(steps_path=steps_path,
                                                features_names=features,
                                                scenarios_names=scenarios,
                                                tags=tags),
        mode=TRAIN_MODE,
        logging_level=logging_level,
        render=render,
        fps_limit=fps,
        do_logs=tensorboard,
        comm_service=ComGameEngineRest(host="127.0.0.1", port=8080)
    )

    test_manager.test_all(features_path)


@cli.command()
@click.option("--iterations", "-i", default=None, help="Number of iterations to test the game.")
@click.option("--features", "-f", default=None, help="List of features to test.")
@click.option("--scenarios", "-s", default=None, help="List of scenarios to test.")
@click.option("--tags", default=None, help="Tags of the features to test.")
@click.option("--log", is_flag=True, help="Log the game.")
@click.option("--debug", is_flag=True, help="Debug debug level.")
@click.option("--info", is_flag=True, help="Info debug level.")
@click.option("--features_path", default=None, type=click.Path(), help="Path of the ./features folder.")
@click.option("--steps_path", default=None, type=click.Path(), help="Path of the ./steps folder.")
def play(debug, features_path, steps_path, iterations, info, log,
         features,
         scenarios,
         tags):
    if iterations is not None:
        iterations = int(iterations)
    else:
        raise Exception("Number of iterations must be set")

    # change start method to fork to avoid errors with multiprocessing
    # Windows does not support the fork start method
    if platform.system() != "Windows":
        set_start_method('fork')

    logging_level = get_debug_level(debug, info)
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging_level)

    mode = TRAIN_MODE

    test_manager = HumanPlayingTestManager(
        feature_strategy=GherkinFeatureStrategy(steps_path=steps_path,
                                                features_names=features,
                                                scenarios_names=scenarios,
                                                tags=tags),
        mode=TEST_MODE,
        logging_level=logging_level,
        iterations=iterations,
        render=True,
        comm_service=ComGameEngineRest(host="127.0.0.1", port=8080)
    )

    test_manager.test_all(features_path)


@cli.command()
@click.option("--features_path", default=None, type=click.Path(), help="Path of the ./features folder.")
@click.option("--steps_path", default=None, type=click.Path(), help="Path of the ./steps folder.")
@click.option("--features", "-f", default=None, help="List of features to test.")
@click.option("--scenarios", "-s", default=None, help="List of scenarios to test.")
@click.option("--tags", default=None, help="Tags of the features to test.")
def list(features, scenarios, tags, features_path="./", steps_path="./"):
    test_manager = DummyTestManager(
        feature_strategy=GherkinFeatureStrategy(steps_path=steps_path, features_names=features,
                                                scenarios_names=scenarios, tags=tags),
        comm_service=ComGameEngineRest(host="127.0.0.1", port=8080))

    test_manager.list_features(features_path)



from __future__ import annotations

from typing import SupportsFloat, Any, Tuple, Dict

import gymnasium as gym
from gymnasium import Space
from gymnasium.core import RenderFrame, ActType, ObsType

from xumes.modules.reinforcement_learning.agent_trainer import AgentTrainer


class GymAdapter(gym.Env):

    def __init__(self,
                 training_service: AgentTrainer,
                 observation_space: Space[ObsType],
                 action_space: Space[ActType],
                 ):
        self._trainer = training_service
        self.observation_space = observation_space
        self.action_space = action_space
        self._first_reset = True

    def reset(
            self,
            *,
            seed: int | None = None,
            options: dict[str, Any] | None = None,
    ) -> tuple[ObsType, dict[str, Any]]:
        self._trainer.do_reset()
        if self._first_reset:
            self._first_reset = False
        else:
            self._trainer.episode_finished()
        return self._trainer.get_obs(), {}

    def step(self, action: ActType) -> Tuple[ObsType, SupportsFloat, bool, bool, Dict[str, Any]]:
        obs = self._trainer.push_actions_and_get_obs(action)
        reward = self._trainer.get_reward()
        terminated = self._trainer.get_terminated()
        return obs, reward, terminated, False, {}

    def render(self) -> RenderFrame | list[RenderFrame] | None:
        return None



from typing import List

from xumes.communication.i_com_game_instance import IComGameInstance


class Behavior:
    pass


class GameInstanceService:
    """
    The `TestRunner` class is a central component of Xumes. It manages communication between communication service,
    the execution of the game itself, and external events that can modify the game state.

    Attributes:
        communication_service (IComGameInstance): An object responsible for communication with other the training service.

    Methods:
        run_communication_service(): Starts the communication service thread.
        run_test_runner(run_func): Starts the game loop if this is the main thread. `run_func` is the game loop function to execute.
        run(): Executes the game by starting both the communication service and the game loop.
        run_render(): Similar to `run()`, but runs the game loop with rendering.
        stop(): Stops both threads currently running.
        wait(): The first method executed in the game loop. It allows the game to wait for an event sent by the training service.
        update_event(event): Method used to accept external modifications to the game, such as reset. `event` represents an external event that can modify the game state.
    """

    def __init__(self,
                 communication_service: IComGameInstance,
                 ):
        self.communication_service = communication_service
        self.is_finished = False

    def run(self, port: int):
        self.communication_service.init_socket(port)

    def stop(self):
        self.communication_service.stop_socket()

    def finish(self):
        self.communication_service.push_dict({"event": "stop"})
        self.communication_service.get_int()

    def push_actions_and_get_state(self, actions: List, methods):
        data = {"event": "action", "inputs": actions, "methods": methods}
        # print(data)
        self.communication_service.push_dict(data)
        return self.communication_service.get_dict()

    def get_state(self):
        self.communication_service.push_dict({"event": "get_state"})
        return self.communication_service.get_dict()

    def get_steps(self):
        self.communication_service.push_dict({"event": "get_steps"})
        return self.communication_service.get_dict()

    def push_args(self, args):
        self.communication_service.push_dict({"event": "args", "args": args})
        self.communication_service.get_int()

    def reset(self):
        self.communication_service.push_dict({"event": "reset"})
        self.communication_service.get_int()



import logging
import traceback

import multiprocess

from xumes import Agent
from xumes.core.modes import TEST_MODE
from xumes.test_automation.test_manager import TestManager
from xumes.modules.reinforcement_learning.vec_stable_baselines_trainer import VecStableBaselinesTrainer


class VecRLTestManager(TestManager):

    def _run_scenarios(self, feature, scenario_datas, active_processes):

        vec_sb_trainers = VecStableBaselinesTrainer()

        model_path = None
        trainers_info = {}

        behaviors: Agent = []

        for scenario in scenario_datas:

            test_runner = scenario_datas[scenario].test_runner
            when_result = test_runner.when()
            if len(when_result) > 1:
                raise Exception("Only one when step is allowed")

            behavior: Agent = when_result[next(iter(when_result))]
            behavior.set_mode(self._mode)
            behavior.set_test_runner(test_runner)

            if trainers_info == {}:
                trainers_info = behavior.get_trainer_info()

            if not model_path:
                model_path = behavior.get_path()

            vec_sb_trainers.add_trainer(behavior.get_trainer())

            # Add behavior to the list in order to free the resources after the training if error occurs
            behaviors.append(behavior)

        def run(_path, nb_processes, vec_trainers, behaviors, do_log=None):
            if not _path:
                _path = "./models/" + feature.name

            logging_path = None
            if do_log:
                logging_path = _path + "/logs"

            try:
                if self._mode == TEST_MODE:
                    vec_trainers.load(_path + "/best_model")
                    vec_trainers.play()
                else:
                    vec_trainers.train(save_path=_path, logs_path=logging_path, logs_name=feature.name, **trainers_info)
            except SystemExit or KeyboardInterrupt:
                logging.warning("Training interrupted by user.")
            except Exception as e:
                logging.error(f"Error occurs while {self._mode}: {e}")
                logging.error(traceback.format_exc())
            finally:
                vec_trainers.free()

                for agent in behaviors:
                    agent.test_runner().finished()

                nb_processes.value -= 1

                exit(0)

        process = multiprocess.Process(target=run, args=(model_path, active_processes, vec_sb_trainers, behaviors, self.do_logs))
        process.start()
        feature.processes.append(process)
        active_processes.value += 1



from typing import Generic, TypeVar

ST = TypeVar("ST")


class GameElementState(Generic[ST]):
    """
        The `GameElementState` class represents the state of a game element.

        Attributes:
            _state: The state object representing the game element state.
            _type: The type of the game element.
            _name: The name of the game element.

        Methods:
            state: Property representing the game element state.
            name: Property representing the name of the game element.
            type: Property representing the type of the game element.
    """
    def __init__(self, name: str, element_type: str, state_obj: ST):
        self._state: ST = state_obj
        self._type = element_type
        self._name = name

    @property
    def state(self) -> ST:
        return self._state

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self):
        return self._type





class AssertionResult:

    def __init__(self, passed: bool, fail_message: str, actual, expected):
        self.passed = passed
        self.fail_message = fail_message
        self.actual = actual
        self.expected = expected




from xumes.core.errors.running_ends_error import RunningEndsError
from xumes.test_automation.behavior import Behavior
from xumes.test_automation.test_manager import TestManager


def play(behavior):
    try:
        behavior.test_runner().reset()
        while True:
            behavior.test_runner().push_action_and_get_state([])

            if behavior.terminated():
                try:
                    behavior.test_runner().episode_finished()
                    behavior.test_runner().reset()
                except RunningEndsError:
                    break
    except Exception as e:
        print("Error in manual playing: ", e)
    finally:
        behavior.test_runner().finished()
        exit(0)


class HumanPlayingTestManager(TestManager):

    def _run_scenarios(self, feature, scenario_datas, active_processes):
        reversed_scenario_datas = list(scenario_datas.keys())
        for scenario in reversed_scenario_datas:
            feature = scenario.feature
            test_runner = scenario_datas[scenario].test_runner

            when_result = test_runner.when()
            if len(when_result) > 1:
                raise Exception("Only one when step is allowed")

            behavior: Behavior = when_result[next(iter(when_result))]
            behavior.set_mode(self._mode)
            behavior.set_test_runner(test_runner)

            play(behavior)



from abc import abstractmethod
from typing import Optional


class ITrainer:

    @abstractmethod
    def train(self, save_path: str = None, eval_freq: int = 10000, logs_path: Optional[str] = None, logs_name: Optional[str] = None, previous_model_path: Optional[str] = None):
        """
        Implementation of the training algorithm.
        """
        raise NotImplementedError

    @abstractmethod
    def save(self, path: str):
        raise NotImplementedError

    @abstractmethod
    def load(self, path: str):
        raise NotImplementedError

    @abstractmethod
    def play(self, timesteps: Optional[int] = None):
        """
        Use the algorithm not in training mode.
        :param test_automation: The test runner to use.
        :param timesteps: Number maximum of step (action to perform).
        """
        raise NotImplementedError

    @abstractmethod
    def free(self):
        """
        Free the resources.
        """
        raise NotImplementedError



from typing import TypeVar

from xumes.entity.game_element_state import GameElementState

SW = TypeVar("SW")


class IGameElementStateConverter:
    """
      The `IGameElementStateConverter` interface defines the method for converting a state element
      to a game element state.

      Methods:
          convert: Converts a state element to a game element state.
    """
    def convert(self, state_wrapper: SW) -> GameElementState:  # TODO add type to state_wrapper
        """
        Converts a state element to a game element state.
        A game element state is easier to work with in order the build entities.
        :param state_wrapper: element of the game state list.
        :returns: A `GameElementState` object representing the converted game element state.
        """
        raise NotImplementedError



import logging
from typing import Optional, List

from stable_baselines3.common.callbacks import EvalCallback
# noinspection PyUnresolvedReferences
from stable_baselines3.common.vec_env import SubprocVecEnv, DummyVecEnv, VecEnvWrapper

from xumes.core.errors.running_ends_error import RunningEndsError
from xumes.modules.reinforcement_learning.i_trainer import ITrainer
from xumes.modules.reinforcement_learning.stable_baselines_trainer import StableBaselinesTrainer


class VecStableBaselinesTrainer(ITrainer):

    def __init__(self):
        self._vec_env = None
        self._trainers: List[StableBaselinesTrainer] = []
        self._envs = []
        self._first_trainer = None
        self.model = None

    def add_trainer(self, trainer: StableBaselinesTrainer):
        self._trainers.append(trainer)
        if self._first_trainer is None:
            self._first_trainer = trainer
        self._envs.append(lambda: trainer.env)

    def make(self):
        self._vec_env = SubprocVecEnv(self._envs, "fork")
        # self._vec_env = DummyVecEnv(self._envs)

    def train(self, save_path: str = None, eval_freq: int = 10000, logs_path: Optional[str] = None,
              logs_name: Optional[str] = None, previous_model_path: Optional[str] = None):
        if self._first_trainer is None:
            raise Exception("No training services added")

        if self._vec_env is None:
            self._vec_env = DummyVecEnv(self._envs)

        algorithm = self._first_trainer.algorithm
        algorithm_type = self._first_trainer.algorithm_type
        total_timesteps = self._first_trainer.total_timesteps

        eval_callback = None
        if save_path:
            eval_callback = EvalCallback(self._vec_env, best_model_save_path=save_path,
                                         log_path=save_path, eval_freq=eval_freq,
                                         deterministic=True, render=False)

        if previous_model_path:
            self.model = algorithm(algorithm_type, self._vec_env, verbose=1, tensorboard_log=logs_path).load(
                previous_model_path, env=self._vec_env).learn(
                total_timesteps,
                callback=eval_callback,
                tb_log_name=logs_name
            )
        else:
            self.model = algorithm(algorithm_type, self._vec_env, verbose=1, tensorboard_log=logs_path).learn(total_timesteps, callback=eval_callback, tb_log_name=logs_name)

    def save(self, path: str):
        self.model.save(path)

    def free(self):
        if self._vec_env is not None:
            self._vec_env.close()
            self.model = None

    def load(self, path: str):
        if self._first_trainer is None:
            raise Exception("No training services added")

        if self._vec_env is None:
            self._vec_env = DummyVecEnv(self._envs)

        algorithm = self._first_trainer.algorithm
        algorithm_type = self._first_trainer.algorithm_type

        self.model = algorithm(algorithm_type, self._vec_env, verbose=1).load(path, env=self._vec_env)

    def play(self, timesteps: int = None):

        class InferenceWrapper(VecEnvWrapper):
            def __init__(self, env):
                super(InferenceWrapper, self).__init__(env)
                self.training = False

            def reset(self):
                return self.venv.reset()

            def step_async(self, a):
                self.venv.step_async(a)

            def step_wait(self):
                return self.venv.step_wait()

        _envs = InferenceWrapper(self._vec_env)
        obs = _envs.reset()

        active_envs = [True] * len(self._envs)

        def step():
            nonlocal obs
            actions, _ = self.model.predict(obs)
            for i in range(len(next(iter(obs.values())))):
                if active_envs[i]:
                    try:
                        single_action = actions[i]
                        single_obs, rewards, done, terminated, info = _envs.envs[i].step(single_action)

                        # Update obs
                        for key, val in single_obs.items():
                            obs[key][i] = val

                        if done or terminated:
                            _envs.envs[i].reset()
                    except RunningEndsError:
                        logging.info(f"Received stop signal for environment {i}. Closing environment.")
                        active_envs[i] = False
                        _envs.envs[i].close()

        if not timesteps:
            while any(active_envs):
                step()
        else:
            for _ in range(timesteps):
                step()
